Index: src/org/armedbear/lisp/sort.lisp
===================================================================
--- src/org/armedbear/lisp/sort.lisp	(revision 13848)
+++ src/org/armedbear/lisp/sort.lisp	(working copy)
@@ -36,7 +36,7 @@
 (defun sort (sequence predicate &rest args &key key)
   (sequence::seq-dispatch sequence
     (sort-list sequence predicate key)
-    (quick-sort sequence 0 (length sequence) predicate key)
+    (quicksort sequence 0 (1- (length sequence)) predicate (or key #'identity))
     (apply #'sequence:sort sequence predicate args)))
 
 (defun stable-sort (sequence predicate &rest args &key key)
@@ -182,29 +182,46 @@
         (if (eq list-1 (cdr head))
             (return list-1))))))
 
-;;; From ECL.
-(defun quick-sort (seq start end pred key)
-  (unless key (setq key #'identity))
-  (if (<= end (1+ start))
-      seq
-      (let* ((j start) (k end) (d (elt seq start)) (kd (funcall key d)))
-        (block outer-loop
-          (loop (loop (decf k)
-                  (unless (< j k) (return-from outer-loop))
-                  (when (funcall pred (funcall key (elt seq k)) kd)
-                    (return)))
-            (loop (incf j)
-              (unless (< j k) (return-from outer-loop))
-              (unless (funcall pred (funcall key (elt seq j)) kd)
-                (return)))
-            (let ((temp (elt seq j)))
-              (setf (elt seq j) (elt seq k)
-                    (elt seq k) temp))))
-        (setf (elt seq start) (elt seq j)
-              (elt seq j) d)
-        (quick-sort seq start j pred key)
-        (quick-sort seq (1+ j) end pred key))))
+;;;
+;;; QUICKSORT
+;;; - the pivot is a middle point
+;;; - sorts the smaller partition first
+;;;
+(defun quicksort (vector start end predicate key)
+  (declare (type fixnum start end)
+	   (type function predicate key))
+  (if (< start end)
+      (let* ((i start)
+	     (j (1+ end))
+	     (p (+ start (ash (- end start) -1)))
+	     (d (aref vector p))
+	     (kd (funcall key d)))
+	(rotatef (aref vector p) (aref vector start))
+	(block outer-loop
+	  (loop
+	    (loop 
+	      (unless (> (decf j) i) (return-from outer-loop))
+	      (when (funcall predicate 
+			     (funcall key (aref vector j)) kd)
+		(return)))
+	    (loop 
+	      (unless (< (incf i) j) (return-from outer-loop))
+	      (unless (funcall predicate
+			       (funcall key (aref vector i)) kd)
+		(return)))
+	    (rotatef (aref vector i) (aref vector j))))
+	(setf (aref vector start) (aref vector j)
+	      (aref vector j) d)
+	(if (< (- j start) (- end j))
+	    (progn
+	      (quicksort vector start (1- j) predicate key)
+	      (quicksort vector (1+ j) end predicate key))
+	    (progn
+	      (quicksort vector (1+ j) end predicate key)
+	      (quicksort vector start (1- j) predicate key))))
+      vector))
 
+
 ;;; From ECL. Should already be user-extensible as it does no type dispatch
 ;;; and uses only user-extensible functions.
 (defun merge (result-type sequence1 sequence2 predicate
